/*
 *  Copyright 2013 University of Nantes, IRCCyN lab.
 *
 *  All Rights Reserved.
 *
 *  Florent Autrusseau.
 *  start : April/2013.
 *  end   : ...........
 *
 *  Questions and comments to:
 *       <mailto:florent.autrusseau@univ-nantes.fr>
 *       <http://www.irccyn.ec-nantes.fr/~autrusse/>
 */

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>
#include <signal.h>
#include <string.h>
#include <assert.h>

#include <it/mat.h>

#include <gsl/gsl_cdf.h>

#include "mpfit.h"

#define FALSE	0
#define TRUE	1


/* This is the private data structure which contains the data points
   and their uncertainties */
struct vars_struct {
  double *x;
  double *y;
  double *ey;
};


/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
/* Simple routine to print the fit results */
void printresult(double *x, double *xact, mp_result *result) 
{
  int i;

  if ((x == 0) || (result == 0)) return;
  printf("  CHI-SQUARE = %f    (%d DOF)\n", 
	 result->bestnorm, result->nfunc-result->nfree);
  printf("        NPAR = %d\n", result->npar);
  printf("       NFREE = %d\n", result->nfree);
  printf("     NPEGGED = %d\n", result->npegged);
  printf("     NITER = %d\n", result->niter);
  printf("      NFEV = %d\n", result->nfev);
  printf("\n");
  if (xact) {
    for (i=0; i<result->npar; i++) {
      printf("  P[%d] = %f +/- %f     (ACTUAL %f)\n", 
	     i, x[i], result->xerror[i], xact[i]);
    }
  } else {
    for (i=0; i<result->npar; i++) {
      printf("  P[%d] = %f +/- %f\n", 
	     i, x[i], result->xerror[i]);
    }
  }
    
}

/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
int gaussfunc(int m, int n, double *p, double *dy, double **dvec, void *vars)
{
  int i;
  struct vars_struct *v = (struct vars_struct *) vars;
  double *x, *y, *ey;
  double xc, sig2;

  x = v->x;
  y = v->y;
  ey = v->ey;

  sig2 = p[3]*p[3];

  for (i=0; i<m; i++) {
    xc = x[i]-p[2];
    dy[i] = (y[i] - p[1]*exp(-0.5*xc*xc/sig2) - p[0])/ey[i];
  }

  return 0;
}

/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
/* Test harness routine, which contains test gaussian-peak data */
int testgaussfit(double *x, double *y, int length, double p[4])
{
  double ey[length];
  double perror[4];			   /* Returned parameter errors */
  mp_par pars[4];			   /* Parameter constraints */         
  int i;
  struct vars_struct v;
  int status;
  mp_result result;
  p[0] = 0.0;
  p[1] = 1.0;
  p[2] = 1.0;
  p[3] = 1.0;       /* Initial conditions */

  memset(&result,0,sizeof(result));      /* Zero results structure */
  result.xerror = perror;

  memset(pars,0,sizeof(pars));        /* Initialize constraint structure */
  /* No constraints */

  for (i=0; i<length; i++) ey[i] = 0.5;

  v.x = x;
  v.y = y;
  v.ey = ey;

  /* Call fitting function for 'length' data points and 4 parameters (no
     parameters fixed) */
  status = mpfit(gaussfunc, length, 4, p, pars, 0, (void *) &v, &result);

//  printf("*** testgaussfit status = %d\n", status);
//  printresult(p, pactual, &result);

  return 0;
}

/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
ivec Grubbs_test_fn(vec datavec, double alpha)
{
	int 		i, j, k, l, len;
	vec 		datavec_b, tmp, tmp2;
	ivec 		idx;
	double 		tcrit=0.0, zcrit=0.0, meanval=0.0, maxval=0.0, tn=0.0, critval=0.0, sdval=0.0;
	int			tmpi, maxvalpos, nb_nan, outlier;

	l = vec_length(datavec);

	datavec_b = vec_new(l);
	tmp = vec_new(l);
	tmp2 = vec_new(l);

	vec_copy(datavec_b, datavec);

	outlier = 1;
	len=l;

	k=1;
	while(outlier == 1)	{
		nb_nan=0;
		for (i=0;i<l;i++)
			if (isnan(datavec_b[i]))
				nb_nan++;
		len = l-nb_nan;
		tmp = vec_new_alloc (len, len);

		j=0;
		for (i=0;i<l;i++)	{
			if (!isnan(datavec_b[i]))	{
				tmp[j] = datavec_b[i];
				j++;
			}
			else
				continue;
		}

		meanval = vec_mean(tmp);
		vec_copy(tmp2,tmp);
		vec_decr(tmp2, meanval);
		vec_abs(tmp2);
		maxvalpos = vec_max_index(tmp2);
		maxval = tmp[maxvalpos];
		sdval = sqrt(vec_variance(tmp));
		tn = fabs((maxval - meanval)/sdval);

		tcrit = gsl_cdf_tdist_Pinv ( alpha/(2*l), l-2);
		zcrit = (l-1)/sqrt(l)*(sqrt(pow(tcrit,2)/(l-2+pow(tcrit,2))));
		critval = zcrit;

		if (tn > critval)
			outlier = 1;
		else
			outlier = 0;

		if (outlier ==1)	{
			for (i=0;i<l;i++)	{
				if (datavec[i] == maxval)	{
					tmpi = i;
				}
			}
			datavec_b[tmpi] = NAN;
		}
	}

	nb_nan=0;
	for (i=0;i<l;i++)	{
		if (isnan(datavec_b[i]))
			nb_nan++;
	}
	idx = ivec_new (nb_nan);

	j=0;
	for (i=0;i<l;i++)
		if (isnan(datavec_b[i]))	{
			idx[j] = i;
			j++;
		}

	vec_delete(datavec);
	vec_delete(datavec_b);
	vec_delete(tmp);
	vec_delete(tmp2);

	return(idx);
}

/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
mat load_TXT(char nom[256], int nb_x, int nb_y)
{

  char filename[256];
  int i, j, a;
  mat TIM;
  FILE *fptr;

  sprintf(filename,"%s",nom);

  if ((fptr = fopen(filename,"r")) == NULL)
	{ printf("Problem while opening file %s\n",filename);exit(1);}

  TIM = mat_new (nb_x, nb_y);
  for (i = 0 ; i < nb_x ; i++)
  {
  	for (j = 0 ; j < nb_y ; j++)
  	{
    	a = fscanf(fptr,"%lf\t",&TIM[i][j]);
	    if (a<0) { printf("load_TXT: Problem reading %s\n",filename); exit(1); }
  	}
  	a = fscanf(fptr,"\n");
  }

  fclose(fptr);

return(TIM);
}

/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
double XCorr(double *x, double *y, int n, int delay)
{
	int i,j,k,d,MaxD;
	double mx, my;
	double diff_x, diff_y;
	double sx, sy, sxy;
	double denom, r;
	double *CrossCorr;

	mx = 0.0;
	my = 0.0;
	diff_x = 0.0;
	diff_y = 0.0;

	//  Compute averages :
	for (i = 0; i < n; i++)
		mx += (double) x[i];
	mx = mx / (n);

	for (i = 0; i < n; i++)
		my += (double) y[i];
	my = my / (n);


	CrossCorr = (double *) calloc (3 * n, sizeof (double));


	sx = 0;
	sy = 0;
	for (i = 0; i < n; i++)
	{
		sx += (x[i] - mx) * (x[i] - mx);
		sy += (y[i] - my) * (y[i] - my);
	}
	denom = sqrt (sx * sy);

	k = 0;
	MaxD = (int) n;
	for (d = -MaxD; d < MaxD; d++)
	{
		sxy = 0;
		for (i = 0; i < n; i++)
		{
			j = i + d;
			if (j < 0 || j >= n)
				continue;
			sxy += (x[i] - mx) * (y[j] - my);
		}
		r = sxy / denom;
		CrossCorr[k] = r;
		k++;
	}

	return(CrossCorr[(int)(k/2)]);

}


/*----------------------------------------------------*/
/*                    function                        */
/*----------------------------------------------------*/
double get_gauss_fit(mat data, mat histo)
{
	double 		*xd, *yd;
	double 		*ydata, *yfit;
	int 		size_h, size_w;
	int 		i, j, k, l;
	double 		MinXcorr=1.0, MaxXcorr=-1.0;
	double		StepXcorr, AmplXcorr;
	double		kf;
	double		max_hist=0.0;
	double 		p[4];
	double 		corrcoef=0.0;
	double 		x_hist[256], y_hist[256];
	vec 		x_histo;
	vec 		y_histo;
	vec			Fit;


	size_w = mat_width(data);
	size_h = mat_height(data);


	xd = (double *) calloc (size_h*size_w, sizeof(double));
	yd = (double *) calloc (size_h*size_w, sizeof(double));
	ydata = (double *) calloc (256, sizeof(double));
	yfit = (double *) calloc (256, sizeof(double));

	for (i=0;i<size_h;i++)
		for (j=0;j<size_w;j++)  {
			if (data[i][j] >= MaxXcorr)	{
				MaxXcorr = data[i][j];
				k=i;
				l=j;
			}
			if (data[i][j] <= MinXcorr)	{
				MinXcorr = data[i][j];
			}
		}


	AmplXcorr = MaxXcorr - MinXcorr;
	StepXcorr = AmplXcorr/255.0;
	x_histo = vec_new_zeros(255);
	y_histo = vec_new_zeros(256);
	Fit = vec_new_zeros(256);

	l=0;
	for (kf=MinXcorr;kf<MaxXcorr;kf+=StepXcorr)	{
     	histo[l][0] = kf;
     	l++;
	}
	l=0;

	for (kf=MinXcorr;kf<MaxXcorr;kf+=StepXcorr)	{
		for (i=0;i<size_h-1;i++)	{
			for (j=0;j<size_w-1;j++)  {
      			if ((data[i][j] > kf) && (data[i][j] <= kf+StepXcorr)) {
					histo[l][1] = histo[l][1]+1;
				}
			}
		}  l++;
	}

    for (i=0;i<256;i++){
      x_histo[i] = histo[i][0];
      y_histo[i] = histo[i][1];
      x_hist[i] = (double)(i);
      y_hist[i] = y_histo[i];
      if (y_hist[i] > max_hist)
      	max_hist = y_hist[i];
    }

	for (i=0;i<256;i++)	{
		xd[i] = histo[i][0];
		yd[i] = y_hist[i];
	}

	i = testgaussfit(xd, yd, 256, p);

	printf("\nGaussian parameters => p: %lf %lf %lf %lf\n",p[0], p[1], p[2], p[3]);

	for (i=0;i<256;i++)	{
		Fit[i] = p[1] * exp(-(pow((xd[i]-(-p[2])),2))/(2*(pow(p[3],2))));
	}

    for (i=0;i<255;i++){
      histo[i][2] = Fit[i];
    }

    for (i=0;i<256;i++)	{
    	ydata[i] = histo[i][1];
    	yfit[i] = histo[i][2];
    }

	corrcoef = XCorr(ydata, yfit, 256, 1);
//	printf("corrcoef: %lf\n",corrcoef);

	free(xd);
	free(yd);
	free(ydata);
	free(yfit);
	vec_delete(x_histo);
	vec_delete(y_histo);
	vec_delete(Fit);

    return(corrcoef);
}

/**************************************************************************************************/
/**************************************************************************************************/
int main(int argc, char **argv)
{
	int 		i;
	char 		filename[256];
	char 		ch_size_w[10], ch_size_h[10], ch_alpha[10];
	int 		size_w, size_h, l;
	mat 		data;
	mat			histo;
	ivec 		idx;
	vec			outliers;
	vec			datavec;
	int			nb_nan;
	double		alpha=0.0;
	double 		corr=0.0;


	if (argc== 5)	{
	    sprintf(filename,"%s",argv[1]);
	    sprintf(ch_size_h,"%s",argv[2]);
	    size_h = atoi(ch_size_h);
	    sprintf(ch_size_w,"%s",argv[3]);
	    size_w = atoi(ch_size_w);
	    sprintf(ch_alpha,"%s",argv[4]);
	    alpha = atof(ch_alpha);
	}
	else	{
		printf("wrong number of arguments...\n");
		printf("example:\n ./Grubbs_test file.txt SizeX SizeY alpha\n");
		printf("example:\n ./Grubbs_test file.txt 64 64 0.05\n");
		exit(0);
	}

	l = size_h*size_w;

	//---   ALLOCATE MEMORY   ---//
	data = mat_new(size_h, size_w);
	datavec = vec_new(l);

	//---   READ DATA FILE   ---//
	data = load_TXT(filename, size_h, size_w);

	//---   CHECK IF GAUSSIAN DISTRIBUTION   ---//
	histo = mat_new_zeros(256, 3);

	corr = get_gauss_fit(data, histo);

	printf("\ncorrcoef: %lf\n\n",corr);
	mat_delete(histo);

	//---   GRUBBS TEST   ---//
	data = _mat_transpose (data);
	datavec = mat_to_vec(data);

	idx = Grubbs_test_fn(datavec, alpha);

	nb_nan= ivec_length(idx);
	outliers = vec_new(nb_nan);

	//---   PRINT OUTLIERS   ---//
	printf("---------------------------");
	printf("\nOutliers from Grubbs test\n");
	printf("---------------------------\n");
	printf("idx: \n");
	for (i=0;i<nb_nan;i++)
		printf("%d\n",idx[i]);

	printf("\noutliers: \n");
	for (i=0;i<nb_nan;i++)	{
		outliers[i] = datavec[(int)(idx[i])];
		printf("%lf\n",outliers[i]);
	}
	printf("\n");

	//---   FREE MEMORY   ---//
  	mat_delete(data);
  	vec_delete(outliers);
  	ivec_delete(idx);

	return (0);

}
